﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using System.Windows.Forms;
using System.Net;
using System.Xml;
using System.Diagnostics;
using System.IO;
using System.Runtime.InteropServices;


namespace TwMon {

    //1tweetのデータ構造
    public struct Tweet {
        public Int64 id;
        public DateTime time;
        public string name, screen_name, text;
    };

    public partial class MainForm : Form {
        
        //iniファイル読み込み用のAPI宣言
        [DllImport("kernel32")]
        private static extern int GetPrivateProfileString(string section,
            string key, string def, StringBuilder retVal, int size, string filePath);

        //リストビューのちらつき防止用
        [DllImport("user32.dll")]
        public static extern int SendMessage(IntPtr hWnd, int msg, int wParam, IntPtr lParam);

        //TwitterのURLテンプレート
        const string APIURL_LIST = "http://api.twitter.com/1/{0}/lists/{1}/statuses.xml?";
        const string APIURL_USER = "http://api.twitter.com/1/statuses/user_timeline.xml?screen_name={0}&";

        const int PERAPI_TWEETS = 20;   //１回で問い合わせるtweets
        const int MAX_TWEETS = 1000;    //格納する最大tweets

        int FirstRequest;           //初回問い合わせ数
        int RequestPeriod;          //問い合わせ間隔(秒)
        string TwitterUser, TwitterList; //ターゲットのユーザ名・リスト名
        string TwitterApiUrl;         //問い合わせURL
        string HighlightID = "";      //ハイライトするID
        Int64 LatestID = 0;           //最新のID

        WebClient TwitterWC = new WebClient();
        Thread WatchThread;
        bool ThreadStop = false;


        //iniファイルの読み込み
        public string IniReadValue(string inifile, string section, string key, string defvalue) {
            StringBuilder sb = new StringBuilder(255);
            int i = GetPrivateProfileString(section, key, defvalue, sb,
                                            255, inifile);
            return sb.ToString().Trim();
        }

        private bool GetIniFile() {
            string exefile = Application.ExecutablePath;
            string inifile = Path.GetDirectoryName(exefile) + "\\"
                + Path.GetFileNameWithoutExtension(exefile) + ".ini";
            
            FirstRequest = int.Parse(IniReadValue(inifile,"TwMon", "FIRST_REQUEST", "5"));
            if(FirstRequest > 10) FirstRequest = 10;

            RequestPeriod = int.Parse(IniReadValue(inifile,"TwMon", "REQUEST_PERIOD", "60"));
            if (RequestPeriod < 30) RequestPeriod = 30;

            TwitterUser = IniReadValue(inifile,"TwMon", "USER", "");
            TwitterList = IniReadValue(inifile,"TwMon", "LIST", "");

            if (TwitterUser == "") {
                MessageBox.Show("ユーザー名が設定されていません。終了します。");
                return false;
            }

            //URL設定
            if (TwitterList == "")
                TwitterApiUrl = string.Format(APIURL_USER, TwitterUser);
            else
                TwitterApiUrl = string.Format(APIURL_LIST, TwitterUser, TwitterList);

            return true;
        }

        //起動時の最初の処理、iniファイルを読めたらフォームを作成
        public MainForm() {
            if (!GetIniFile())
                Environment.Exit(0);

            InitializeComponent();

        }

        private void MainForm_Load(object sender, EventArgs e) {
            //リストビューのちらつき防止
            int styles = (int)SendMessage(listTweet.Handle, 0x1037, 0, (IntPtr)0);
            styles |= 0x10000;
            SendMessage(listTweet.Handle, 0x1036, 0, (IntPtr)styles);

            this.Text = string.Format("TwMon {0}    user:{1}  list:{2}",
                "v1.1", TwitterUser, TwitterList);

            //監視スレッド開始
            WatchThread = new Thread(new ThreadStart(TweetMonitor));
            WatchThread.Start();
        }


        //APIでデータを取得
        private List<Tweet> GetTweets(string msg, string param) {
            List<Tweet> tws = new List<Tweet>();
            string str = "";

            //APIでデータ取得
            ShowStatus(msg + ": Twitter API発行中...");
            try {
                byte[] data = TwitterWC.DownloadData(TwitterApiUrl + param);
                str = Encoding.UTF8.GetString(data);
            }
            catch(WebException we) {
                ShowError(we.Message);
                return tws;
            }
            ShowError("");

            //XMLの解析
            ShowStatus(msg + ": 解析中...");
            XmlDocument xdoc = new XmlDocument();
            xdoc.LoadXml(str);

            //XPathでstatusの抽出
            XmlNodeList nl = xdoc.SelectNodes("/statuses/status");
            foreach (XmlNode node in nl) {
                Tweet tw = new Tweet();
                tw.id          = Int64.Parse(node["id"].InnerText);
                tw.name        = node["user"]["name"].InnerText;
                tw.screen_name = node["user"]["screen_name"].InnerText;
                tw.text        = node["text"].InnerText;
                tw.time        = DateTime.ParseExact(node["created_at"].InnerText,
                    "ddd MMM dd HH':'mm':'ss zz00 yyyy",
                    System.Globalization.DateTimeFormatInfo.InvariantInfo);
                tws.Add(tw);
            }
            ShowStatus("");
            ShowError(string.Format("{0} 件のTweetsを取得しました。",tws.Count));
            return tws;
        }

        //マルチスレッドでコントロールを操作するためdelegate作成
        private delegate void AddTweetsDelegate(List<Tweet> tws, bool add_front);
        private void AddTweets(List<Tweet> tws, bool add_front) {
            if (ThreadStop) return;
            Invoke(new AddTweetsDelegate(AddTweets_sub), tws, add_front);
        }

        //取得したデータを、内部のリスト及びlistviewに追加
        private void AddTweets_sub(List<Tweet> tws, bool add_front) {

            //最新のID番号を記憶
            if(tws.Count > 0 && tws[0].id > LatestID)
                LatestID = tws[0].id;

            listTweet.BeginUpdate();

            if(add_front) {
                //頭に追加するために、逆順にしておく
                tws.Reverse();
                //既存の項目の色調整
                foreach(ListViewItem item in listTweet.Items) {
                    //真っ白な行についたらそこで終了
                    if (item.BackColor.B == 0xFF)
                        break;
                    //1/5ずつ増やしていく
                    //ID欄以外の名前部分も色をそろえておく
                    item.BackColor = item.SubItems[0].BackColor = Color.FromArgb(0xff, 0xff, item.BackColor.B + 0x33);
                }
            }

            foreach(Tweet tw in tws) {
                ListViewItem item = new ListViewItem();
                string status_url = string.Format("http://twitter.com/{0}/status/{1}",tw.screen_name,tw.id.ToString());
                item.Text = tw.time.ToString();
                item.SubItems.Add(tw.screen_name);
                item.SubItems.Add(tw.text);
                item.SubItems.Add(tw.id.ToString());
                item.SubItems.Add(status_url);
                item.SubItems.Add(string.Format("時刻：{0}    ID：{1}  ユーザ名：{2}\nURL ： {3}\n\n{4}",
                    tw.time.ToString("yyyy/MM/dd HH:mm:ss"),tw.screen_name, tw.name, status_url, tw.text));

                //ハイライト周りの設定
                item.SubItems[1].BackColor = item.SubItems[2].BackColor = Color.LightBlue;
                item.UseItemStyleForSubItems = (item.SubItems[1].Text != HighlightID);
                
                if(add_front) {
                    //色をつける
                    item.BackColor = item.SubItems[0].BackColor = Color.FromArgb(0xff, 0xff, 0);
                    listTweet.Items.Insert(0, item);
                }
                else {
                    item.BackColor = item.SubItems[0].BackColor =  Color.FromArgb(0xff, 0xff, 0xff);
                    listTweet.Items.Add(item);
                }
            }

            //多すぎる分は消す
            while(listTweet.Items.Count > MAX_TWEETS)
                listTweet.Items.RemoveAt(MAX_TWEETS);

            listTweet.EndUpdate();

        }

        void TweetMonitor(){
            //Twitterの書き込み監視スレッド
            int Wait;
            List<Tweet> tws;

            //初期化が終わるまでは、とりあえず連続してデータを取得
            tws = GetTweets(string.Format("最近{0}件のデータを取得します",PERAPI_TWEETS),"");
            AddTweets(tws,false);

            //先頭、指定ページ分のデータを取得
            for (int i = 1; i < FirstRequest && tws.Count == PERAPI_TWEETS; i++) {
                tws = GetTweets(string.Format("{0}～{1}件目のデータを取得します",
                    PERAPI_TWEETS * i + 1, PERAPI_TWEETS * (i + 1)),
                    "max_id=" + (tws[PERAPI_TWEETS-1].id-1).ToString());
                AddTweets(tws, false);
            }
            
            Wait = RequestPeriod;

            //以下監視のメインループ
            while(!ThreadStop){
                ShowStatus(string.Format("次回の更新は {0} 秒後です。",Wait));
                Wait--;

                if(Wait==0) {
                    //現在保持している最新のIDより新しいものを取りにに行く
                    string since = "since_id=" + (LatestID + 1).ToString();
                    tws = GetTweets(string.Format("新着データを取りに行きます"),since);
                    AddTweets(tws,true);

                    //またしばらく待つ                        
                    Wait = RequestPeriod;
                }

                Thread.Sleep(1000);
            }

        }



        //ステータス表示用のデリゲート
        private void ShowStatus(string str) {
            Action<string> act = delegate(string str2) { toolStripStatusLabel1.Text = str2; };
            if(ThreadStop) return;
            Invoke(act, str);
        }
        private void ShowError(string str) {
            Action<string> act = delegate(string str2) { toolStripStatusLabel2.Text = str2; };
            if (ThreadStop) return;
            Invoke(act, str);
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e) {
            //スレッドも終了
            ThreadStop = true;
            WatchThread.Abort();
            WatchThread.Join();
        }


        //行選択時処理
        private void listTweet_SelectedIndexChanged(object sender, EventArgs e) {

            if (listTweet.SelectedItems.Count<1)
                return;

            //上の表示部分
            labelTweet.Links.Clear();
            labelTweet.Text = string.Format("[{0}/{1}]  {2}",listTweet.SelectedIndices[0]+1,listTweet.Items.Count,
                listTweet.SelectedItems[0].SubItems[5].Text);

            //リンクをはる
            Regex reg = new Regex(@"https?://[-_.!~*'()a-zA-Z0-9;/?:@&=+$,%#]+");
            foreach(Match m in reg.Matches(labelTweet.Text)){
                labelTweet.Links.Add(m.Index,m.Length,m.Groups[0].Value);
            }

            //同一のIDの色をつける
            HighlightID = listTweet.SelectedItems[0].SubItems[1].Text;
            listTweet.BeginUpdate();
            //IDが一致するものはSubItemの装飾設定を有効にするという横着
            foreach(ListViewItem item in listTweet.Items)
                item.UseItemStyleForSubItems = (item.SubItems[1].Text != HighlightID);
            listTweet.EndUpdate();


        }

        private void listTweet_MouseClick(object sender, MouseEventArgs e) {
            //右クリック時、データをクリップボードにコピー
            if (e.Button == MouseButtons.Right) {
                ListViewItem item = listTweet.GetItemAt(e.X, e.Y);
                if (item != null) {
                    Clipboard.SetText(item.SubItems[5].Text);
                    toolStripStatusLabel2.Text = "つぶやきをクリップボードにコピーしました。";
                }
            }
        }

        private void listTweet_MouseDoubleClick(object sender, MouseEventArgs e) {
            //ダブルクリック時、ブラウザで対象のつぶやきを開く
            ListViewItem item = listTweet.GetItemAt(e.X, e.Y);

            //URLは事前に仕込み済み
            if (e.Button == MouseButtons.Left && item != null)
                Process.Start(item.SubItems[4].Text);
        }

        private void labelTweet_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e) {
            //URLクリック時にブラウザ起動
            Process.Start(e.Link.LinkData.ToString());
        }

    }
    static class Program {
        /// <summary>
        /// アプリケーションのメイン エントリ ポイントです。
        /// </summary>
        [STAThread]
        static void Main() {
            Application.EnableVisualStyles();
            Application.SetCompatibleTextRenderingDefault(false);
            Application.Run(new MainForm());
        }
    }
}